#!/bin/bash
#
# Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
#
# This software product is a proprietary product of NVIDIA CORPORATION &
# AFFILIATES (the "Company") and all right, title, and interest in and to the
# software product, including all associated intellectual property rights, are
# and shall remain exclusively with the Company.
#
# This software product is governed by the End User License Agreement
# provided with the software product.
#

# This script is used to initiate IPsec connection using strongSwan between two DPUs.

####################
## Configurations ##
####################

script_folder_path=`dirname "$0"`
script_path="$0"

os_name=`cat /etc/os-release | grep -w 'NAME=' | cut -d '"' -f2 | cut -d ' ' -f1`
pciconf=`ls --color=never /dev/mst/ | grep --color=never '^m.*f0$' | cut -c 3-`
mlnx_bf_conf_file="/etc/mellanox/mlnx-bf.conf"
back_up_folder="/var/tmp/east_west_backup"
back_up_file_cert="${back_up_folder}/backup_cert.txt"
back_up_file_addr="${back_up_folder}/backup_addr.txt"
back_up_file_port_num="${back_up_folder}/backup_portnum.txt"

# Define certificates and key folder according to the OS.
if [[ "$os_name" == "Ubuntu" ]]; then
        CERT_FOLDER="/etc/swanctl/x509"
        KEY_FOLDER="/etc/swanctl/private"
        CACERT_FOLDER="/etc/swanctl/x509ca"
elif [[ "$os_name" == "CentOS" ]]; then
        CERT_FOLDER="/etc/strongswan/swanctl/x509"
        KEY_FOLDER="/etc/strongswan/swanctl/private"
        CACERT_FOLDER="/etc/strongswan/swanctl/x509ca"
fi

RED="\033[0;31m"
GREEN="\033[0;32m"
NOCOLOR="\033[0;0m"

######################
## Helper functions ##
######################

function usage() {
        cat << HEREDOC
        usage: `readlink -f "$script_path"` [COMMAND] [PARAMS...]
        Configures IPsec connection between two BlueFields with HW full offload.
        PARAMS can be be either configured in JSON parameters file or passed on the command line.
        Example using $params_file_path:
                On one of the BlueFields: $script_path --side=r --json=east_west_overlay_encryption_params.json
                On the second BlueField (after the first one finishes): $script_path --side=i \
--json=east_west_overlay_encryption_params.json
                To destroy the connection:
                $script_path -d|--destroy
        Example passing the PARAMS:
                On one of the BlueFields: $script_path --side=r --initiator_ip_addr=192.168.50.1 \
--receiver_ip_addr=192.168.50.2 --port_num=0 --auth_method=psk --preshared_key=swordfish
                On the second BlueField (after the first one finishes): $script_path --side=i \
--initiator_ip_addr=192.168.50.1 --receiver_ip_addr=192.168.50.2 --port_num=0 --auth_method=psk \
--preshared_key=swordfish
                To destroy the connection:
                $script_path -d|--destroy
        Command and parameters:
        -d, --destory                         Destroys the IPsec connection and reverts the configurations.
                                              If not using the parameters file, then port_num must be
                                              passed with this command.
        -h, --help                            Displays this help text and exit.
        --side=SIDE                           Side of the connection, can be r (meaning reciever, used when
                                              running on the first BlueField), or i (meaning initiator,
                                              used when running on the second BlueField). Note that this
                                              flag can't be configured in the parameters file and must be
                                              passed with the command line.
        -j=PARAMS_FILE, --json=PARAMS_FILE    JSON parameters file that includes the relevant parameters. When using
                                              this flag there is no need to pass any other parameters other than the
                                              side or destroy command.
                                              Note: the JSON file must follow the template as in the JSON
                                              file the came with this script (east_west_overlay_encryption_params.json).
        --initiator_ip_addr=IP_ADDRESS        The initiator's port IP address.
        --receiver_ip_addr=IP_ADDRESS         The receiver's port IP address.
        --port_num=NUM                        The number of the port, can be 0 or 1.
        --auth_method=METHOD                  The authentication method, can be psk (meaning pre-shared key
                                              authentication method), ssc (meaning self-signed
                                              certificates authentication method) or CA (meaning
                                              certificate authority certificates).
        --preshared_key=PSK                   The pre-shared key (relevant if auth_method is psk).
        --initiator_cert_path=CERT_PATH       The initiator's certificate path (relevant if auth_method is
                                              scc or ca).
        --receiver_cert_path=CERT_PATH        The receiver's certificate path (relevant if auth_method is
                                              scc or ca).
        --initiator_key_path=KEY_PATH         The initiator's private-key path (relevant if side is
                                              initiator and auth_method is scc or ca).
        --receiver_key_path=KEY_PATH          The receiver's private-key path (relevant if side is
                                              receiver and auth_method is scc or ca).
        --initiator_cacert_path=CACERT_PATH   The initiator's CA certificate path (relevant if side is
                                              initiator and auth_method is ca).
        --receiver_cacert_path=CACERT_PATH    The receiver's CA certificate path (relevant if side is
                                              receiver and auth_method is ca).
        --initiator_cn=CN                     The common name (CN) of the initiator's certificate (relevant
                                              if side is receiver and auth_method is ca).
        --receiver_cn=CN                      The common name (CN) of the receiver's certificate (relevant
                                              if side is initiator and auth_method is ca).
HEREDOC

        exit $1
}

# This function outputs error message that is passed to this function as argument.
function err() {
        m="ERROR: $1"
        echo -e "${RED}$m$NOCOLOR"
}

# Delete the backup files and the certificates and exit the script.
function exit_and_clean() {
        # Delete certificates and key if exist.
        if [[ -f $back_up_file_cert ]]; then
                while IFS= read -r line; do
                        rm -f $line > /dev/null 2>&1
                done < $back_up_file_cert
        fi

        # Remove the backup folder
        rm -rf $back_up_folder

        # Exit
        exit $1
}

# Delete from ovs bridges ports that we use.
function delete_prev_bridges() {
        echo "Deleting vxlan-br${port_num} (if exists) and deleting $PF and $PF_REP from all bridges."
        # Delete vxlan-br${port_num} bridge if exists
        ovs-vsctl del-br vxlan-br${port_num} > /dev/null 2>&1
        # Delete PF and PF_REP ports from all other bridges
        for bridge in `ovs-vsctl show |  grep -oP 'Bridge \K\w[^\s]+'`; do
                ovs-vsctl del-port $bridge $PF > /dev/null 2>&1
                ovs-vsctl del-port $bridge $PF_REP > /dev/null 2>&1
        done
}

# This function outputs the MTU of the device that is passed as argument to the function.
function get_mtu() {
        local device=$1
        local device_mtu=`ifconfig $device | grep mtu | awk '{print $4}'`
        echo $device_mtu
}

# Configure IPsec over OVS VXLAN bridge according to authentication method.
function ovs_config() {
        # Delete ovs bridges and ports that we use.
        delete_prev_bridges

        ifconfig $PF $LOCAL_IP/24 up
        ifconfig $PF_REP up

        # Set the MTU of PF to be 50 more of the PF_REP's MTU to account for the size of VXLAN headers.
        local pf_mtu=$(get_mtu $PF_REP)
        let pf_mtu+=50
        echo "Setting the MTU of $PF to $pf_mtu (50 more than the MTU of $PF_REP) to account for the size of VXLAN headers."
        ifconfig $PF mtu $pf_mtu up

        # Enable tc offloading
        echo "Updating hw-tc-offload to $PF."
        ethtool -K $PF hw-tc-offload on
        # Disable host PF as the port owner
        mlxprivhost -d /dev/mst/mt${pciconf} --disable_port_owner r

        # Start openvswitch according to the operating system.
        if [[ "$os_name" == "Ubuntu" ]]; then
                service openvswitch-switch start
        elif [[ "$os_name" == "CentOS" ]]; then
                service openvswitch start
        fi

        # Start OVS IPsec.
        systemctl start openvswitch-ipsec.service

        # Configure OVS VXLAN IPsec bridge.
        bride_name=vxlan-br${port_num}
        ovs-vsctl add-br $bride_name
        ovs-vsctl add-port $bride_name $PF_REP
        ovs-vsctl set Open_vSwitch . other_config:hw-offload=true

        # Configure tunnel type, port key and destination port.
        local tunnel_type="vxlan"
        local port_key=100
        # destination port of vxlan tunnel is always 4789.
        local dst_port=4789

        # Choose authentication method.
        if [[ "$auth_method" == "psk" ]]; then
                echo "Configuring IPsec using pre-shared key."
                ovs_config_psk $tunnel_type $port_key $dst_port
        elif [[ "$auth_method" == "ssc" ]]; then
                echo "Configuring IPsec using self-signed certificates."
                ovs_config_ssc $tunnel_type $port_key $dst_port
        elif [[ "$auth_method" == "ca" ]]; then
                echo "Configuring IPsec using CA-signed Certificate."
                ovs_config_ca $tunnel_type $port_key $dst_port
        fi

        # Backup IP addresses to delete the IPsec rules in the future
        if [[ ! -f $back_up_file_addr ]]; then
                if [ ! -d $back_up_folder ]; then
                        mkdir -p $back_up_folder
                fi
                touch $back_up_file_addr
        fi
        echo "$LOCAL_IP" >> $back_up_file_addr
        echo "$REMOTE_IP" >> $back_up_file_addr

        # Backup port number to flush the port in the future
        touch $back_up_file_port_num
        echo "$port_num" >> $back_up_file_port_num
}

# Configure IPsec over OVS VXLAN bridge using pre-shared key.
# Tunnel type, port key and dst_port are passed as arguments to the funtion (in that order).
function ovs_config_psk() {
        # Configure arguments.
        local tunnel_type=$1
        local port_key=$2
        local dst_port=$3

        # Add IPsec rule to the bridge.
        ovs-vsctl add-port $bride_name tun -- \
                set interface tun type=$tunnel_type \
                options:local_ip=$LOCAL_IP options:remote_ip=$REMOTE_IP \
                options:key=$port_key options:dst_port=$dst_port \
                options:psk=$preshared_key

        # Restart openvswitch according to the operating system.
        if [[ "$os_name" == "Ubuntu" ]]; then
                service openvswitch-switch restart
        elif [[ "$os_name" == "CentOS" ]]; then
                service openvswitch restart
        fi
        ovs-vsctl show
}

# Configure IPsec over OVS VXLAN bridge using self-signed certificates.
# Tunnel type, port key and dst_port are passed as arguments to the funtion (in that order).
function ovs_config_ssc() {
        # Configure arguments.
        local tunnel_type=$1
        local port_key=$2
        local dst_port=$3

        # Configure certificates and key.
        cp $LOCAL_CERT $REMOTE_CERT $CERT_FOLDER
        cp $KEY $KEY_FOLDER

        # Update variables.
        LOCAL_CERT="${CERT_FOLDER}/`basename $LOCAL_CERT`"
        REMOTE_CERT="${CERT_FOLDER}/`basename $REMOTE_CERT`"
        KEY="${KEY_FOLDER}/`basename $KEY`"

        # Add IPsec rule to the bridge.
        ovs-vsctl set Open_vSwitch . \
                other_config:certificate="$LOCAL_CERT" \
                other_config:private_key="$KEY"

        ovs-vsctl add-port $bride_name ${tunnel_type}p0 -- set interface ${tunnel_type}p0 type=$tunnel_type \
                options:local_ip=$LOCAL_IP options:remote_ip=$REMOTE_IP \
                options:key=$port_key options:dst_port=$dst_port \
                options:remote_cert="$REMOTE_CERT"

        # Restart openvswitch according to the operating system.
        if [[ "$os_name" == "Ubuntu" ]]; then
                service openvswitch-switch restart
        elif [[ "$os_name" == "CentOS" ]]; then
                service openvswitch restart
        fi

        # Backup Certificates and key path so we delete them when reverting IPsec configurations.
        if [[ ! -f $back_up_file_cert ]]; then
                if [ ! -d $back_up_folder ]; then
                        mkdir -p $back_up_folder
                fi
                touch $back_up_file_cert
        fi
        echo "$LOCAL_CERT" >> $back_up_file_cert
        echo "$REMOTE_CERT" >> $back_up_file_cert
        echo "$KEY" >> $back_up_file_cert

        ovs-vsctl show
}

# Configure IPsec over OVS VXLAN bridge using certificate authority (CA) certificates.
# Tunnel type, port key and dst_port are passed as arguments to the funtion (in that order).
function ovs_config_ca() {
        # Configure arguments.
        local tunnel_type=$1
        local port_key=$2
        local dst_port=$3

        # Configure certificates and key.
        cp $LOCAL_CERT $REMOTE_CERT $CERT_FOLDER
        cp $KEY $KEY_FOLDER
        cp $CACERT $CACERT_FOLDER

        # Update variables.
        LOCAL_CERT="${CERT_FOLDER}/`basename $LOCAL_CERT`"
        REMOTE_CERT="${CERT_FOLDER}/`basename $REMOTE_CERT`"
        KEY="${KEY_FOLDER}/`basename $KEY`"
        CACERT="${CACERT_FOLDER}/`basename $CACERT`"

        # Add IPsec rule to the bridge.
        ovs-vsctl set Open_vSwitch . \
                other_config:certificate="$LOCAL_CERT" \
                other_config:private_key="$KEY" \
                other_config:ca_cert="$CACERT"

        ovs-vsctl add-port $bride_name ${tunnel_type}p0 -- set interface ${tunnel_type}p0 type=$tunnel_type \
                options:local_ip=$LOCAL_IP options:remote_ip=$REMOTE_IP \
                options:key=$port_key options:dst_port=$dst_port \
                options:remote_name=$REMOTE_CN

        # Restart openvswitch according to the operating system.
        if [[ "$os_name" == "Ubuntu" ]]; then
                service openvswitch-switch restart
        elif [[ "$os_name" == "CentOS" ]]; then
                service openvswitch restart
        fi

        # Backup Certificates and key paths so we delete them when reverting IPsec configurations.
        if [[ ! -f $back_up_file_cert ]]; then
                if [ ! -d $back_up_folder ]; then
                        mkdir -p $back_up_folder
                fi
                touch $back_up_file_cert
        fi
        echo "$LOCAL_CERT" >> $back_up_file_cert
        echo "$REMOTE_CERT" >> $back_up_file_cert
        echo "$KEY" >> $back_up_file_cert
        echo "$CACERT" >> $back_up_file_cert

        ovs-vsctl show
}

# Configure IPsec offload mode to "full" if argument is 1, Configure IPsec offload mode to "none" if argument is 0.
# Note: This function restarts the ib driver.
function config_offload() {
        # Disable regex temporarily.
        echo "Disabling regex temporarily to configure IPsec HW offload mode."
        systemctl stop mlx-regex

        if (( $1 == 1 )) ; then
                echo "Configuring IPsec HW offload mode to \"full\"."
                 # Replace "no" with "yes" in conf_file.
                search_string="IPSEC_FULL_OFFLOAD=\"no\""
                replace_string="IPSEC_FULL_OFFLOAD=\"yes\""
                ipsec_wanted_mode="full"
        elif (( $1 == 0 )) ; then
                echo "Configuring IPsec HW offload mode to \"none\"."
                # Replace "yes" with "no" in conf_file.
                search_string="IPSEC_FULL_OFFLOAD=\"yes\""
                replace_string="IPSEC_FULL_OFFLOAD=\"no\""
                ipsec_wanted_mode="none"
        fi

        if  grep -Fxq "$replace_string" $mlnx_bf_conf_file; then
                : # Do nothing
        elif grep -Fxq "$search_string" $mlnx_bf_conf_file; then
                sed -i "s/$search_string/$replace_string/" $mlnx_bf_conf_file
        fi

        # Reload driver to apply changes.
        echo "Restarting ib driver (/etc/init.d/openibd restart)."
        /etc/init.d/openibd restart

        # Check if ipsec_mode is correct.
        if [[ -f /sys/class/net/p${port_num}/compat/devlink/ipsec_mode ]]; then
                ipsec_current_mode=$(cat /sys/class/net/p${port_num}/compat/devlink/ipsec_mode)
                diff  <(echo $ipsec_wanted_mode) <(echo $ipsec_current_mode) > /dev/null 2>&1
                if (( $? == 0 )); then
                        echo "Configuration succeeded. IPsec HW offload mode is \"$ipsec_current_mode\"."
                else
                        err "IPSec offload configuration failed. IPsec HW offload mode is \"$ipsec_current_mode\", mode should be \"$ipsec_wanted_mode\"."
                        # Re-enable regex.
                        echo "Re-enabling regex."
                        systemctl restart mlx-regex
                        echo "Exiting."
                        # Exit and remove backup folder
                        exit_and_clean 1
                fi
        else
                err "Can't show IPsec offload mode."
                # Re-enable regex.
                echo "Re-enabling regex."
                systemctl restart mlx-regex
                # Exit and remove backup folder
                echo "Exiting."
                exit_and_clean 1
        fi

        # Re-enable regex.
        echo "Re-enabling regex."
        systemctl restart mlx-regex
}

# This function delete IPsec states and policies according to the two IP addresses (LOCAL and REMOTE) that are
# passed to this function as parameters.
function delete_ipsec_rules() {
        # Delete relevant policies
        for spi in `ip xfrm state | grep -A1 "src $1 dst $2" | grep spi | awk '{ print $4 }'`; do
                ip xfrm state delete src $1 dst $2 proto esp spi $spi
        done
        for spi in `ip xfrm state | grep -A1 "src $2 dst $1" | grep spi | awk '{ print $4 }'`; do
                ip xfrm state delete src $2 dst $1 proto esp spi $spi
        done

        # Delete relevant states
        ip xfrm policy | grep "src ${1}/32 dst ${2}/32" | while read policy; do
                ip xfrm policy delete $policy dir out
        done
        ip xfrm policy | grep "src ${2}/32 dst ${1}/32" | while read policy; do
                ip xfrm policy delete $policy dir in
        done


}

# This function reverts IPsec rules (if they exist), disables IPsec full offload mode,
# and deletes local certificates and key files (if they exist).
function revert_ipsec() {
        echo "Reverting IPsec configurations..."
        # Get the relevant IP addresses and delete IPsec rules
        if [[ -f $back_up_file_addr ]]; then
                LOCAL_IP=`sed '1q;d' $back_up_file_addr`
                REMOTE_IP=`sed '2q;d' $back_up_file_addr`
                # Delete the IPsec rules.
                delete_ipsec_rules $LOCAL_IP $REMOTE_IP
        else
                err "Can't delete IPsec rules because couldn't retrieve IP addresses from the backup file: $back_up_file_addr."
        fi


        sudo systemctl start NetworkManager
        # Get the port number and flush the port
        if [[ -f $back_up_file_port_num ]]; then
                port_num=`cat $back_up_file_port_num`
                # Flush the port
                sudo ip addr flush dev p${port_num}
        else
                err "Can't flush the port that was used because couldn't retrieve port number from the backup file: $back_up_file_port_num."
        fi
        sudo rm -f /etc/openvswitch/conf.db

        # Disable host restriction
        mlxprivhost -d /dev/mst/mt${pciconf} p

        # Restart openvswitch according to the operating system.
        if [[ "$os_name" == "Ubuntu" ]]; then
                sudo /etc/init.d/openvswitch-switch restart
        elif [[ "$os_name" == "CentOS" ]]; then
                systemctl restart openvswitch
                systemctl enable openvswitch
        fi

        # Disable IPsec full offload.
        config_offload 0

        # Delete certificates and key if exist.
        if [[ -f $back_up_file_cert ]]; then
                while IFS= read -r line; do
                        rm -f $line > /dev/null 2>&1
                        done < $back_up_file_cert
        fi
        sudo rm -rf $back_up_folder

        echo "Reverted IPsec configurations."
        echo "Don't forget to revert the configurations on the other machine (if relevant)."
}


# This functions outputs the number of SADs
function read_sad_counter() {
        local counter=`ip xfrm state count | awk '{ print $3 }'`
        echo $counter
}

# This functions checks if the IPsec connection is established:
#       Comparing the initial SAD counter (which is passed as argument to the function) with current SAD counter.
#       Checking the IPsec SAD entires if they are correct for this connection.
function validate_ipsec() {
        # Define initial and current SAD states.
        local initial_sad_count=$1
        local current_sad_count=$(read_sad_counter)

        # Check if the current SAD counter is bigger than the initial SAD counter.
        if [ "$current_sad_count" -le "$initial_sad_count" ]; then
                err "IPsec connection failed, check parameters and configurations and then try again."
                revert_ipsec
                exit 1
        fi

        # Also check if the entries are correct for this connection.
        ip x s | grep "src $LOCAL_IP dst $REMOTE_IP" > /dev/null 2>&1
        if (( $? !=0 )); then
                err "IPsec connection failed, check parameters and configurations and then try again."
                revert_ipsec
                exit 1
        fi
        ip x s | grep "src $REMOTE_IP dst $LOCAL_IP" > /dev/null 2>&1
        if (( $? !=0 )); then
                err "IPsec connection failed, check parameters and configurations and then try again."
                revert_ipsec
                exit 1
        fi
        echo -e "${GREEN}IPsec connection is now established. You can now send encrypted" \
                "traffic between ${LOCAL_IP} and ${REMOTE_IP}.$NOCOLOR"
}

# This function recieves as parameters the devices required for the IPsec connection and checks if they are avaiable.
# If no devices are found the function tries to fix the issue by restarting the ib driver.
# If the fix didn't help, we exit the script.
function check_devices() {
        local PF=$1
        local PF_REP=$2
        echo "Checking devices $PF and $PF_REP."
        ifconfig $PF up > /dev/null 2>&1
        if (( $? != 0 )); then
                echo "$PF device not found, trying to fix."
                # Reload ib driver to find port.
                echo "Restarting ib driver (/etc/init.d/openibd restart)."
                /etc/init.d/openibd restart
                ifconfig $PF up > /dev/null 2>&1
                if (( $? != 0 )); then
                        err "Fix didn't work, $PF device not found, exiting."
                        # Exit and remove backup folder
                        exit_and_clean 1
                fi
        fi

        ifconfig $PF_REP up > /dev/null 2>&1
        if (( $? != 0 )); then
                echo "$PF_REP device not found, trying to fix."
                # Reload ib driver to find port.
                echo "Restarting ib driver (/etc/init.d/openibd restart)."
                /etc/init.d/openibd restart
                ifconfig $PF_REP up > /dev/null 2>&1
                if (( $? != 0 )); then
                        err "Fix didn't work, $PF_REP device not found, exiting."
                        sudo ip addr flush dev p${port_num}
                        # Exit and remove backup folder
                        exit_and_clean 1
                fi
        fi
}

# This function checks if the relevant certificates and keys exist.
# The certificates and keys are passed as arguments to the function.
# The CA certificate should be always passed last.
# If the relevant certificates and keys don't exist then we exit the script.
function check_certificates_key() {
        echo "Checking certificates and keys."
        for ((i = 1; i < $#; i++ )); do
                arg=${!i}
                if [[ ! -f "$arg" ]]; then
                        err "$arg does not exist. Exiting."
                        # Exit and remove backup folder
                        exit_and_clean 1
                fi
        done

        # If ca mode check if the CA certificate exists.
        # The CA certificate is the last argument passed to the function.
        if [[ "$auth_method" == "ca" ]]; then
                arg=${!#}
                if [[ ! -f "$arg" ]]; then
                        err "$arg does not exist. Exiting."
                        # Exit and remove backup folder
                        exit_and_clean 1
                fi
        fi
}

# This function check's from the initiator's side if there is a valid connection between the initiator and the receiver
# using device that is passed as argument to the function.
function check_connection() {
        # Delete ovs bridges and ports that we use so we can check the connection.
        delete_prev_bridges

        # Give the device IP address to check connection.
        device=$1
        ifconfig $device $LOCAL_IP/24 up > /dev/null 2>&1

        # Check if there's a connection between the initiator and receiver side using PF.
        if [[ "$side" == "initiator" || "$side" == "i" ]]; then
                echo "Checking connection between $LOCAL_IP and $REMOTE_IP..."
                ping -c1 $REMOTE_IP > /dev/null 2>&1
                if (( $? !=0 )); then
                        err "Can't ping between $LOCAL_IP and $REMOTE_IP. Exiting."
                        sudo ip addr flush dev p${port_num}
                        # Exit and remove backup folder
                        exit_and_clean 1
                else
                        echo "Connection is successful."
                fi
        fi
}

# This function parses the parameters from the json file
function parse_json_file() {
        # Check if file exists as absoulte path or relative path
        json_file_relative=${script_folder_path}/${json_file}
        if [[ -f $json_file ]]; then
                :
        elif [[ -f $json_file_relative ]]; then
                json_file=$json_file_relative
        else
                err "JSON file $json_file does not exist. Exiting."
                # Exit and remove backup folder
                exit_and_clean 1
        fi

        echo "Parsing parameters from $json_file..."
        initiator_ip_addr=`grep '"initiator_ip_addr":' $json_file | awk -F "\"" '{ print $4 }'`
        receiver_ip_addr=`grep '"receiver_ip_addr":' $json_file | awk -F "\"" '{ print $4 }'`
        port_num=`grep '"port_num":' $json_file | awk -F "\"" '{ print $4 }'`
        auth_method=`grep '"auth_method":' $json_file | awk -F "\"" '{ print $4 }'`
        preshared_key=`grep '"preshared_key":' $json_file | awk -F "\"" '{ print $4 }'`
        initiator_cert_path=`grep '"initiator_cert_path":' $json_file | awk -F "\"" '{ print $4 }'`
        receiver_cert_path=`grep '"receiver_cert_path":' $json_file | awk -F "\"" '{ print $4 }'`
        initiator_key_path=`grep '"initiator_key_path":' $json_file | awk -F "\"" '{ print $4 }'`
        receiver_key_path=`grep '"receiver_key_path":' $json_file | awk -F "\"" '{ print $4 }'`
        initiator_cacert_path=`grep '"initiator_cacert_path":' $json_file | awk -F "\"" '{ print $4 }'`
        receiver_cacert_path=`grep '"receiver_cacert_path":' $json_file | awk -F "\"" '{ print $4 }'`
        initiator_cn=`grep '"initiator_cn":' $json_file | awk -F "\"" '{ print $4 }'`
        receiver_cn=`grep '"receiver_cn":' $json_file | awk -F "\"" '{ print $4 }'`
}

# This functions parses the parameters of the script.
function parse_parameters() {
        num_of_params=$#
        for i in "$@"; do
                case $i in
                        -j=*)
                                json_file="${i#*=}"
                                # Move to the next parameter
                                shift
                                ;;
                        --json=*)
                                json_file="${i#*=}"
                                # Move to the next parameter
                                shift
                                ;;
                        --side=*)
                                side="${i#*=}"
                                # Move to the next parameter
                                shift
                                ;;
                        --initiator_ip_addr=*)
                                initiator_ip_addr="${i#*=}"
                                # Move to the next parameter
                                shift
                                ;;
                        --receiver_ip_addr=*)
                                # Move to the next parameter
                                receiver_ip_addr="${i#*=}"
                                shift
                                ;;
                        --port_num=*)
                                port_num="${i#*=}"
                                # Move to the next parameter
                                shift
                                ;;
                        --auth_method=*)
                                auth_method="${i#*=}"
                                # Move to the next parameter
                                shift
                                ;;
                        --preshared_key=*)
                                preshared_key="${i#*=}"
                                # Move to the next parameter
                                shift
                                ;;
                        --initiator_cert_path=*)
                                initiator_cert_path="${i#*=}"
                                # Move to the next parameter
                                shift
                                ;;
                        --receiver_cert_path=*)
                                # Move to the next parameter
                                receiver_cert_path="${i#*=}"
                                shift
                                ;;
                        --initiator_key_path=*)
                                initiator_key_path="${i#*=}"
                                # Move to the next parameter
                                shift
                                ;;
                        --receiver_key_path=*)
                                receiver_key_path="${i#*=}"
                                # Move to the next parameter
                                shift
                                ;;
                        --initiator_cacert_path=*)
                                initiator_cacert_path="${i#*=}"
                                # Move to the next parameter
                                shift
                                ;;
                        --receiver_cacert_path=*)
                                receiver_cacert_path="${i#*=}"
                                # Move to the next parameter
                                shift
                                ;;
                        --initiator_cn=*)
                                initiator_cn="${i#*=}"
                                # Move to the next parameter
                                shift
                                ;;
                        --receiver_cn=*)
                                receiver_cn="${i#*=}"
                                # Move to the next parameter
                                shift
                                ;;
                        -d|--destroy)
                                command="$i"
                                # Move to the next parameter.
                                shift
                                ;;
                        -h|--help)
                                command="$i"
                                # Move to the next parameter.
                                shift
                                ;;
                        *)
                                err "Illegal parameter $i. Run the script with --help for usage."
                                exit 1
                                ;;
                esac
        done

        # Parse the parameters from the json file if it was passed
        if [[ -z $json_file ]]; then
                : # If not set then do nothing
        else
                # If set check if the number of params is legal
                if (( $num_of_params > 2 )); then
                        err "Illegal number of parameters! When running with JSON file, the only other parameter that \
can be passed is the side. Exiting"
                        # Exit and remove backup folder
                        exit_and_clean 1
                fi
                parse_json_file
        fi
}

# This function checks if the parameters passed to the function are set.
# Parameter name is passed with the parameter value in case of an unset parameter.
function check_parameters_set() {
        while [[ $# -gt 0 ]]; do
                # Check if parameter is set
                if [[ -z $2 ]]; then
                        # Get parameter name
                        parameter_name=$1
                        err "Parameter $parameter_name was not set. Run the script with --help for usage."
                        exit 1
                else
                        # Move to the next parameter
                        shift 2
                fi
        done
}

# This function checks and configures the parameters according to the side and auth_method
function configure_parameters() {
        # If command is help or destroy then no need for checking and configuring parameters
        if [[ "$command" == "--help" || "$command" == "-h" || "$command" == "--destroy" || "$command" == "-d" ]]; then
                return 0
        fi

        # Check if the needed parameters were set
        check_parameters_set "side" "$side" "initiator_ip_addr" "$initiator_ip_addr" \
                "receiver_ip_addr" "$receiver_ip_addr" "port_num" "$port_num" "auth_method" "$auth_method"

        # Check if the side is legal
        if [[ "$side" != "initiator" && "$side" != "i" && "$side" != "receiver" && "$side" != "r" ]]; then
                err "Illegal side $side, must be i|initiator|r|receiver. Exiting."
                # Exit and remove backup folder
                exit_and_clean 1
        fi

        # Check the rest of the parameters according to auth_method and side
        if [[ "$auth_method" == "psk" ]]; then
                check_parameters_set "preshared_key" $preshared_key
        elif [[ "$auth_method" == "ssc" ]]; then
                if [[ "$side" == "initiator" || "$side" == "i" ]]; then
                        check_parameters_set "initiator_cert_path" "$initiator_cert_path" \
                                "receiver_cert_path" "$receiver_cert_path" "initiator_key_path" "$initiator_key_path"
                elif [[ "$side" == "receiver" || "$side" == "r" ]]; then
                        check_parameters_set "initiator_cert_path" "$initiator_cert_path" \
                                "receiver_cert_path" "$receiver_cert_path" "receiver_key_path" "$receiver_key_path"
                fi
        elif [[ "$auth_method" == "ca" ]]; then
                if [[ "$side" == "initiator" || "$side" == "i" ]]; then
                        check_parameters_set "initiator_cert_path" "$initiator_cert_path" \
                                "receiver_cert_path" "$receiver_cert_path" "initiator_key_path" "$initiator_key_path" \
                                "initiator_cacert_path" "$initiator_cacert_path" "receiver_cn" "$receiver_cn"
                elif [[ "$side" == "receiver" || "$side" == "r" ]]; then
                        check_parameters_set "initiator_cert_path" "$initiator_cert_path" \
                                "receiver_cert_path" "$receiver_cert_path" "receiver_key_path" "$receiver_key_path" \
                                "receiver_cacert_path" "$receiver_cacert_path" "initiator_cn" "$initiator_cn"
                fi
        else
                # Illegal authentication method
                err "Illegal auth_method $auth_method, must be psk|ssc|ca. Exiting."
                # Exit and remove backup folder
                exit_and_clean 1
        fi

        # Check port number if legal.
        if (( $port_num != 0 && $port_num != 1 )); then
                err "Illegal port_num $port_num, port_num must be 0 or 1. Exiting"
                # Exit and remove backup folder
                exit_and_clean 1
        fi

        # Configure parameters
        PF=p${port_num}
        PF_REP=pf${port_num}hpf
        if [[ "$side" == "initiator" || "$side" == "i" ]]; then
                LOCAL_IP=$initiator_ip_addr
                REMOTE_IP=$receiver_ip_addr
                LOCAL_CERT=$initiator_cert_path
                REMOTE_CERT=$receiver_cert_path
                CACERT=$initiator_cacert_path
                KEY=$initiator_key_path
                REMOTE_CN=$receiver_cn
        elif [[ "$side" == "receiver" || "$side" == "r" ]]; then
                LOCAL_IP=$receiver_ip_addr
                REMOTE_IP=$initiator_ip_addr
                LOCAL_CERT=$receiver_cert_path
                REMOTE_CERT=$initiator_cert_path
                CACERT=$receiver_cacert_path
                KEY=$receiver_key_path
                REMOTE_CN=$initiator_cn
        fi

        # If in "ssc" or "ca" mode then check if the certificates and key exist.
        if [[ "$auth_method" == "ssc" || "$auth_method" == "ca" ]]; then
                check_certificates_key $LOCAL_CERT $REMOTE_CERT $KEY $CACERT
        fi


}

# This function parses and checks the script's parameters.
function get_parameters() {
        # Parse the parameter
        parse_parameters "$@"

        # Configure and check the parameters
        configure_parameters

        if [[ "$command" == "--help" || "$command" == "-h" ]]; then
                usage 0
        elif [[ "$command" == "--destroy" || "$command" == "-d" ]]; then
                revert_ipsec
                exit 0
        fi

        # Check devices needed for the IPsec connection.
        check_devices $PF $PF_REP

        # Check connection using PF.
        check_connection $PF
}

##################
## Script Start ##
##################

# Parse and check script's parameters.
get_parameters "$@"

# Get initial SAD count to compare after the connection is established.
initial_sad_count=$(read_sad_counter)

# Enable IPsec full offload.
config_offload 1

# Configure the OVS VXLAN IPsec bridge.
ovs_config

# Check if the IPsec connection is established by checking if the SAD count increased and if the SAD entries are correct.
if [[ "$side" == "initiator" || "$side" == "i" ]]; then
        echo "Waiting 10 seconds for the IPsec connection to be established."
        sleep 10s
        validate_ipsec $initial_sad_count
fi
